import {
  ApolloClient,
  ApolloLink,
  HttpLink,
  InMemoryCache,
  split,
  ApolloProvider,
  fromPromise,
  FetchResult,
} from "@apollo/client";
import { withScalars } from "apollo-link-scalars";
import { buildClientSchema, IntrospectionQuery } from "graphql";
import { DateTimeResolver } from "graphql-scalars";
import { FC } from "react";
import introspectionResult from "../../generated/graphql.schema.json";
import { onError } from "@apollo/client/link/error";
import { WebSocketLink } from "@apollo/client/link/ws";
import { getMainDefinition, Observable } from "@apollo/client/utilities";
import { useSnackbar } from "notistack";
import { deepOmit } from "../../utils/deep-omit";
import { useMemo } from "react";
import { EventEmitter } from "events";
import { useState } from "react";
import { useAuth } from "../auth/auth.provider";
import { useEffect } from "react";
import { useRef } from "react";
import { setContext } from "@apollo/client/link/context";

const schema = buildClientSchema(
  introspectionResult as unknown as IntrospectionQuery
);

const typesMap = {
  DateTime: DateTimeResolver,
};

const cleanTypeName = new ApolloLink((operation, forward) => {
  if (operation.variables) {
    operation.variables = deepOmit(["__typename"], operation.variables);
  }
  const rt = forward(operation);
  return (
    rt.map?.((data) => {
      return data;
    }) ?? rt
  );
});

export const AppApolloClientProvider: FC = ({ children }) => {
  const { enqueueSnackbar } = useSnackbar();
  const { accessToken, logout } = useAuth();

  const [loggedEventTarget] = useState(() => new EventEmitter());
  const accessTokenRef = useRef(accessToken);
  const logoutRef = useRef(logout);

  useEffect(() => {
    accessTokenRef.current = accessToken;
    if (accessToken) {
      loggedEventTarget.emit("logged", accessToken);
    }
  }, [loggedEventTarget, accessToken]);
  useEffect(() => {
    logoutRef.current = logout;
  }, [logout]);
  const client = useMemo(() => {
    const authLink = onError(
      ({ graphQLErrors, networkError, operation, forward }) => {
        if (graphQLErrors) {
          for (const error of graphQLErrors) {
            if (error.extensions?.code === "401") {
              return fromPromise(
                new Promise<Observable<FetchResult>>((resolve) => {
                  loggedEventTarget.once("logged", (accessToken: string) => {
                    const oldHeaders = operation.getContext().headers;
                    operation.setContext({
                      headers: {
                        ...oldHeaders,
                        authorization: `Bearer ${accessToken}`,
                      },
                    });
                    resolve(forward(operation));
                  });
                  logoutRef.current();
                })
              ).flatMap((v) => v);
            }
          }
        }
        const httpResult = (networkError as any)?.result;
        if (httpResult?.statusCode === 401) {
          return fromPromise(
            new Promise<Observable<FetchResult>>((resolve) => {
              loggedEventTarget.once("logged", (accessToken: string) => {
                const oldHeaders = operation.getContext().headers;
                operation.setContext({
                  headers: {
                    ...oldHeaders,
                    authorization: `Bearer ${accessToken}`,
                  },
                });
                resolve(forward(operation));
              });
              logoutRef.current();
            })
          ).flatMap((v) => v);
        }
      }
    ).concat(
      setContext(() => ({
        headers: {
          authorization: `Bearer ${accessTokenRef.current}`,
        },
      }))
    );
    const errorLink = onError(({ graphQLErrors, networkError }) => {
      if (graphQLErrors) {
        graphQLErrors.forEach((error) => {
          enqueueSnackbar(error.message, {
            variant: "error",
          });
        });
        graphQLErrors.forEach(({ message, locations, path }) => {
          console.error(
            `[GraphQL error]: Message: ${message}, Location: ${locations}, Path: ${path}`
          );
        });
      }
      if (networkError) {
        console.log(`[Network error]: ${networkError}`);
        enqueueSnackbar(networkError.message, {
          variant: "error",
        });
      }
    });
    const wsLink = new WebSocketLink({
      uri: `${window.location.protocol.replace("http", "ws")}//${
        window.location.hostname
      }:${window.location.port}/api/graphql`,
      options: {
        reconnect: true,
      },
    });
    const httpLink = new HttpLink({
      uri: "/api/graphql",
    });
    const splitLink = split(
      ({ query }) => {
        const definition = getMainDefinition(query);
        return (
          definition.kind === "OperationDefinition" &&
          definition.operation === "subscription"
        );
      },
      wsLink,
      httpLink
    );
    const link = ApolloLink.from([
      errorLink,
      authLink,
      withScalars({ schema, typesMap }) as unknown as ApolloLink,
      cleanTypeName,
      splitLink,
    ]);
    const client = new ApolloClient({
      connectToDevTools: true,
      ssrMode: typeof window === "undefined",
      link,
      cache: new InMemoryCache({
        typePolicies: {},
      }),
    });

    return client;
  }, [enqueueSnackbar, loggedEventTarget]);

  return <ApolloProvider client={client}>{children}</ApolloProvider>;
};
